/*
NIXIE TUBE SLOT MACHINE

Be_Zero_Be, Poulsbo WA, USA
Started : December 9, 2023
Release : October 12, 2024

**********************************************************************************
This is my first C++ program.
I have written many thousands of lines of Visual Basic, Assembly and PLC Ladder logic.
I am sure there are more elegant ways to implement this program.
This source code is heavily commented due to frequent cerebral malfunctions.
None the less, here is is. :) 
serial.Print statements are left in as comments for debuging and analysis 
**********************************************************************************

Symbol nixie tubes are used as wheels of a nixie slot machine.
I chose the IN-7 nixie symbol tube because it has 10 unique symbols.
The IN-15A also has 10 unique symbols.
Any symbol or numeric nixie tube can be used.

The actual probability of winning in this arrangement are 10 three of a kind in 1000 possibilites.
That is an average of 1 in 100 spins.
To keep the player interested the odds are reduced to 1 in every "MaxSpins".
The slot machine will payoff at least once in every 12 spins.
The payoff will be 3 to 15 "Coins."
At any time, the random number generator can output an actual winning number sequence.
This is a randon real 1 in 100 BONUS win and has a BONUS payoff of 35 - 50 coins.

If the Nixie Slot Machine is left on and not used for a long period of time,
a symbol can (over time) become dim or burn out (called Cathode Poisoning). 
If the game is on and not played for 20 minutes it will 
"spin the wheels" to protect the nixie tube.
This spin does not effect play in progress, only the display is exercised.

Pressing the PLAY button for five seconds will reset the game to 25 coins.

Define I/O and initialization
We use 12 Arduino digital bits to address the nixie tube decimal decoder chips.
Four digital bits are used to address each of the nixie displays (Left, Mid. Right).
One analog bit is used as a digital input for the "Play Button".
74141/k155id1 decoder address lines are based on PCB layout.
A, B, C, D, Address decoder lines are mapped to Arduino I/O pins. 
*/
#define LeftA 16  // Left 74141/k155id1 decoder address lines.
#define LeftB 18
#define LeftC 9
#define LeftD 17
#define MidA 19  // Middle 74141/k155id1 decoder address lines.
#define MidB 7
#define MidC 6
#define MidD 8
#define RightA 5  //Right 74141/k155id1 decoder address lines.
#define RightB 3
#define RightC 2
#define RightD 4
#define OrangeNeon 14        // orange neon bulb output bit
#define GreenNeon 15         // green neon bulb output bit
#define PlayButton 7         // analog input for Play button
#define latchPin 11          //74HC595 RCLK  pin 12
#define clockPin 12          //74HC595 SRCLK pin 11
#define dataPin 10           //74HC595 SER   pin 14 of first 74HC595
#define CathodeTime 1200000  //20 minute Cathode Poisoning timer (1,200,000 mS)
#define LEDpin 13

#include <EEPROM.h>  // for EEPROM read/write

// GLOBAL DECLARIATIONS

void ShowWinners();
void ShowCoins();
void SpinWheels();
void SaveCathode();
void GameOver();
void PayCoins();

/* Array of 10 winning numbers.
The PCB mapping of the decoder outputs (Q0-Q9)
to the nixie tube pins, determines the 10 winning numbers.
Your winning numbers will be determiined by your PCB layout. 
These numbers represent the output bits (Q0-Q9)
of the thrree 74141/k155id1 nixie driver decoder chips. 
All three Qn output numbers are deliberatly different.
Otherwise the random number generator would have to 
generate three consecutive same digits, possible but not likely. */

byte Winner[30] = {
  //                            Left Mid Right
  //                             U7  U8  U9
  2, 7, 6,  // 276 = "П" symbol  Q6, Q7, Q2
  3, 6, 7,  // 367 = "+" symbol  Q7, Q7, Q3
  7, 3, 2,  // 732 = "m" symbol  Q2, Q3, Q7
  6, 2, 3,  // 623 = "A" symbol  Q3, Q2, Q6
  0, 4, 8,  // 048 = "V" symbol  Q6, Q4, Q0
  1, 9, 0,  // 190 = "M" symbol  Q0, Q9, Q1
  5, 0, 1,  // 501 = "Ω" symbol  Q1, Q0, Q5
  4, 8, 9,  // 489 = "~" symbol  Q9, Q9, Q4
  8, 1, 5,  // 815 = "K" symbol  Q5, Q1, Q8
  9, 5, 4,  // 945 = "-" Symbol  Q4, Q5, Q9
};
// payout per symbol
//                       П, +, m, A, V, M, Ω, ~,  K,  -
int SymbolPayOut[10] = { 3, 4, 5, 6, 7, 8, 9, 10, 12, 15 };

unsigned long Cathode;       // Nixie tube symbol preservation timer.
unsigned long Reset;         // 5 second score reset
unsigned long WriteCounter;  // count how many times we write to the EEPROM
bool toggle;                 // flipper bit.
int Spin;                    // Play counter.
int PayoffSpin;              // Random Spin for payoff.
int PayoffChar;              // Random nixie symbol will be displayed for the win.
int PayoffCoins;             // payoff coins for PayoffSpin
int Win;                     // winning number array pointer
int SaveCharLeft;            // Save the three symbols during "cathode" spin
int SaveCharMid;             // ditto
int SaveCharRight;           // ditto
int PullArm;                 // pull the one armed bandits arm to spin the wheels
int EE_addrCoins;            // EEPROM address of "Coins"
int EE_addrWriteCounter;     // EEPROM address of WriteCount
int EE_Pointer;              // pointer to data in EEPROM
int Coins;                   // coin count

// MAIN PROGRAM LOOP
void loop() {
  /* Analog input (A7) is used for Play button input
  pressing the play button will ground the input to A7
  If PullArm < 5  = analog value button pushed. It will pull down to zero normally
  Check the 20 minute cathode timer. */
  if (millis() > Cathode) SaveCathode();  // Exercise the neons
  Reset = millis() + 5000;                // Score reset to 25 Coins after 5 seconds
  PullArm = analogRead(PlayButton);
  if (PullArm < 5) {
    do {  // hold for five seconds to reset the Coins to 25
      if (millis() > Reset) {
        ShowWinners();
        Coins = 26;  // set to 26 - one will be used to spin the wheels to restart the game
        EE_WriteCoins();
        ShowCoins();
        break;
      }
      PullArm = analogRead(PlayButton);  // confirm play button pressed
    } while (PullArm < 5);
    Spin++;                            // Increment the Spin counter.
    Cathode = millis() + CathodeTime;  // 1,200,000 Reset the Cathode timer every spin.
    Coins--;                           // take the coin for this play
    EE_WriteCoins();                   // update coin count in EEPROM
    ShowCoins();                       // Show the coin count
    SpinWheels();                      // Spin the wheels
  }
  // if Coins goes to zero or > 9999 then game over.
  if ((Coins == 0) || (Coins >= 9999)) GameOver();
}

// get the random numbers for the next game segment.
void GetRandom() {
  PayoffSpin = random(3, 12);                  // A number between  3 and 11 for next payoff.
  PayoffChar = random(0, 10) * 3;              // Array pointer for a payoff symbol between (0 to 9) * 3
  PayoffCoins = SymbolPayOut[PayoffChar / 3];  // PayOff according to symbol
  Spin = 0;                                    // reset the spin counter.
}